/* libexs/list.h */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef LIBEXS_LIST_H
#define LIBEXS_LIST_H

#include <pthread.h>

enum {
	OUR_LIST_BLOCKING = 1,
	OUR_LIST_ALL_FLAGS = OUR_LIST_BLOCKING,
};

/* Implementation in this file based on the two-lock concurrent queue algorithm
 * proposed in
 *   [MS1996] M. Michael and M. Scott. Simple, Fast, and Practical Non-Blocking
 *   and Blocking Concurrent Queue Algorithms, 1996.
 *
 * Concurrent calls to exs_p_list_add_tail and exs_p_list_del_head may be made
 * on the same list without any external synchronization.
 *
 * All of the functions and structures in this file are private to the UNH-EXS
 * implementation and not considered part of the UNH-EXS API, unless otherwise
 * noted. Therefore, these function prototypes and structure definitions may
 * change without any notice. */

struct our_list_head {
	struct our_list_slot *head, *tail;
	struct our_list_slot *empty_head, *empty_tail;
	pthread_mutex_t head_mutex;
	pthread_mutex_t tail_mutex;
	pthread_cond_t empty_cv; /* guarded by tail_mutex */
	unsigned int wait_count; /* guarded by tail_mutex */
	unsigned int flags;
};

struct our_list_slot {
	void *value;
	struct our_list_slot *next;
};

/* Function prototypes */
int
exs_p_list_init(struct our_list_head *list, int flags);

void
exs_p_list_destroy(struct our_list_head *list);

int
exs_p_list_add_tail(struct our_list_head *list, void *value);

void *
exs_p_list_del_head_or_wait(struct our_list_head *list,
		struct timespec *deadline);

void *
exs_p_list_del_head(struct our_list_head *list);

void *
exs_p_list_peek_head(struct our_list_head *list);

int
exs_p_list_splice(struct our_list_head *l1, struct our_list_head *l2);

void
exs_p_list_unblock(struct our_list_head *list);

/* Returns a nonzero value if the list is empty at some time during which this
 * method is called. Due to concurrent access to these lists, however, the
 * return value of this function should not be relied upon. */
static inline int
exs_p_list_empty(struct our_list_head *list)
{
	int ret;
	pthread_mutex_lock(&list->head_mutex);
	ret = (list->head->next == NULL);
	pthread_mutex_unlock(&list->head_mutex);
	return ret;
}

/* Returns the number of elements in the list at some time during which this
 * method is called. Due to concurrent access to these lists, however, the
 * return value of this function should not be relied upon. */
static inline int
exs_p_list_size(struct our_list_head *list)
{
	struct our_list_slot *ptr;
	int ret = 0;

	pthread_mutex_lock(&list->tail_mutex);
	pthread_mutex_lock(&list->head_mutex);
	for (ptr = list->head->next; ptr != NULL; ptr = ptr->next) {
		++ret;
	}
	pthread_mutex_unlock(&list->head_mutex);
	pthread_mutex_unlock(&list->tail_mutex);
	return ret;
}

/* A wrapper around pthread_mutex_(timed)?lock that calls the appropriate
 * function based on whether deadline is NULL or non-NULL. */
static inline int
our_mutex_timedlock(pthread_mutex_t *mutex, struct timespec *deadline)
{
	int ret;
	if (deadline) {
		ret = pthread_mutex_timedlock(mutex, deadline);
	} else {
		ret = pthread_mutex_lock(mutex);
	}
	return ret;
}

/* A wrapper around pthread_cond_(timed)?wait that calls the appropriate
 * function based on whether deadline is NULL or non-NULL. */
static inline int
our_cv_timedwait(pthread_cond_t *cv, pthread_mutex_t *mutex,
						struct timespec *deadline)
{
	int ret;
	if (deadline) {
		ret = pthread_cond_timedwait(cv, mutex, deadline);
	} else {
		ret = pthread_cond_wait(cv, mutex);
	}
	return ret;
}

#endif  /* LIBEXS_LIST_H */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
