/* our_calloc_free.c */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE
#define _XOPEN_SOURCE 600

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include <our_calloc_free.h>

#define	MAX_RECORDS	1000

struct alloc_record {
	size_t		nmemb;
	size_t		size;
	const char	*fname;
	unsigned int	lineno;
	pthread_t	allocator;
	void		*address;
};


/* table of all currently allocated heap memory */
static struct alloc_record	records[MAX_RECORDS];

/* ptr to first slot in records table that is MAYBE unused */
static struct alloc_record	*first_maybe_unused_record = records;

/* ptr to next slot in records table that starts remaining unused records */
static struct alloc_record	*last_used_plus_1 = records;

/* protect all access to records table and the two boundary pointers */
static pthread_mutex_t		record_mutex = PTHREAD_MUTEX_INITIALIZER;


/*******************************************************************************
 * invariants:
 *
 *	all records preceding first_maybe_unused_record are used (lineno != 0)
 *
 *	all records at or after last_used_plus_1 are unused (lineno == 0)
 *
 * strategy:
 *	try to keep the table densely packed without ever moving a record
 *
 ******************************************************************************/


/* allocate (nmemb * size) bytes of zeroed memory and return pointer to it.
 *
 * info about this allocation is stored in the first currently empty record.
 */
void *
_our_calloc_calloc( size_t nmemb, size_t size, const char *fname,
		   unsigned int lineno )
{
	struct alloc_record	*ptr;
	void			*address;


	/* first try to do the desired allocation, leave if it fails */
	if ((address = calloc(nmemb, size)) == NULL)
		return address;

	/* protect all access to table and pointers into it */
	pthread_mutex_lock(&record_mutex);

	/* find first empty record in which to store this allocation */
	for (ptr = first_maybe_unused_record; ptr != last_used_plus_1; ptr++){
		if (ptr->lineno == 0)
			break;
	}

	if (ptr == &records[MAX_RECORDS]) {
		TRACE_DBG(EXS_LOG_ALLOC,
			"line %u: calloc(%zu, %zu) not recorded, all %u records are full",
			lineno, nmemb, size, MAX_RECORDS);
	} else {/* save in the empty record all info about this allocation */
		ptr->nmemb = nmemb;
		ptr->size = size;
		ptr->fname = fname;
		ptr->lineno = lineno;
		ptr->allocator = pthread_self();
		ptr->address = address;
		if (ptr == last_used_plus_1) {
			/* just filled in the last record in use so far */
			last_used_plus_1++;
		/* just filled in the first known unused record,
		 * so first "maybe" unused record might be the next one
		 */
		first_maybe_unused_record = ptr + 1;
		}
	}

	pthread_mutex_unlock(&record_mutex);

	/* always return the allocated chunk of zeroed-out memory */
	return address;
}	/* _our_calloc_calloc */


/* free the heap memory previously allocated at address.
 *
 * table is searched to find a record of the previous allocation
 *	if not found, a message is printed
 *	if found, that record is marked unused and boundary markers are updated
 *		first_maybe_unused_record will point at first unused record
 *		last_used_plus_1 will point 1 beyond last used record
 */
void
_our_calloc_free( void *address, unsigned int lineno )
{
	struct alloc_record	*ptr;
	struct alloc_record	*first_known_unused_record;


	/* protect all access to table and pointers into it */
	pthread_mutex_lock(&record_mutex);

	/* try to find a record to match this address */
	first_known_unused_record = NULL;
	for (ptr = records; ptr != last_used_plus_1; ptr++) {
		if (ptr->lineno == 0) {
			/* hit an unused slot, remember it if it is first one */
			if (first_known_unused_record == NULL)
				first_known_unused_record = ptr;
		} else if (ptr->address == address) {
			/* found used slot with matching allocation address */
			break;
		}
	}

	if (ptr == last_used_plus_1) {
		/* did not find a record with this address */
		TRACE_ERR("line %u: free address %p that was not recorded",
				lineno, address);
		goto errout;
	} else {/* found a record with this address, free it */
		ptr->lineno = 0;	
		if (first_known_unused_record == NULL) {
			/* just freed what is now the lowest unused record */
			first_maybe_unused_record = ptr;
		} else {/* hit a known unused record before the one we freed */
			first_maybe_unused_record = first_known_unused_record;
		}
		if (ptr == last_used_plus_1 - 1) {
			/* we just freed the highest numbered record in use,
			 * push back the next_to_use limit
			 */
			do {
				last_used_plus_1--;
				ptr--;
			} while (last_used_plus_1 != records
							&& ptr->lineno == 0);
		}
	}

	/* always free the caller's memory */
	TRACE_DBG(EXS_LOG_ALLOC,
		"freeing %p first_maybe_unused_record %p last_used_plus_1 %p",
		address,
		first_maybe_unused_record, last_used_plus_1);
	free(address);
errout:
	pthread_mutex_unlock(&record_mutex);
}	/* _our_calloc_free */

void
_our_calloc_report(FILE *fptr)
{
	unsigned int		count;
	struct alloc_record	*ptr;

	/* report all unfreed allocations */
	pthread_mutex_lock(&record_mutex);
	count = 0;
	for (ptr = records; ptr != last_used_plus_1; ptr++) {
		if (ptr->lineno != 0) {
			count++;
			fprintf(fptr, "%4u: line %u in function %s thread %lu"
				" calloc(%lu, %lu) = %p not freed\n",
				count,
				ptr->lineno,
				ptr->fname,
				ptr->allocator,
				(unsigned long)ptr->nmemb,
				(unsigned long)ptr->size,
				ptr->address);
		}
	}
	fprintf(fptr, "%4u items allocated but not freed\n", count);
	fflush(fptr);
	pthread_mutex_unlock(&record_mutex);

}	/* _our_calloc_report */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
