/* our_calloc_free.h */

#ifndef OUR_CALLOC_FREE_H_
#define OUR_CALLOC_FREE_H_

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define EXS_EXPORT __attribute__((visibility("default")))

#if defined(ALLOW_EXS_STATISTICS)

	enum statistics_counter_type {
		completion_loops,
		good_ibv_poll_cqs,
		empty_ibv_poll_cqs,
		good_ibv_get_cq_events,
		good_completions,
		send_advert_completions,
		send_rdma_write_completions,
		recv_advert_completions,
		recv_rdma_with_imm_completions,
		recv_eof_completions,
		total_setup_completions,
		matched_recv_await,
		unmatched_recv_await,
		send_done_after_sending,
		send_done_after_completion,
		total_setup_sends,
		advert_blocks_sent,
		signaled_advert_blocks_sent,
		unsignaled_advert_blocks_sent,
		signaled_rdma_write_blocks_sent,
		unsignaled_rdma_write_blocks_sent,
		rdma_write_blocks_sent,
		rdma_write_bytes_sent,
		ack_blocks_sent,
		recv_ack_completions,
		eof_blocks_sent,
		extra_ack_blocks,
		direct_rdma_write_blocks_sent,
		indirect_rdma_write_blocks_sent,
		no_credits_on_send_wwi,
		no_credits_on_send_ack,
		no_credits_on_send_advert,
		direct_indirect_switches,
		ibv_reg_mr_calls,
		ibv_dereg_mr_calls,
		max_statistics_counter_type
	};

#define STATISTICS_INCREMENT(arg)	\
	do	{		\
		statistics_counters[arg]++;		\
	} while (0)

#define STATISTICS_DECREMENT(arg)	\
	do	{		\
		statistics_counters[arg]--;		\
	} while (0)

#define STATISTICS_ADD(arg, val)	\
	do	{		\
		statistics_counters[arg] += val;	\
	} while (0)

#else

#define STATISTICS_INCREMENT(arg)	\
	do	{		\
	} while (0)

#define STATISTICS_ADD(arg, val)	\
	do	{		\
	} while (0)

#endif

char *_time_prefix(void);

void
exs_log(int module, int priority,
		const char *file, int line, const char *function,
		const char *format, ...)
		__attribute__((format(printf, 6, 7)));

enum exs_log_priority {
	EXS_LOG_ERROR,
	EXS_LOG_INFO,
	EXS_LOG_DEBUG,
};

enum exs_log_module {
	EXS_LOG_UNSPECIFIED = 1 >> 0,
	EXS_LOG_ALLOC = 1 >> 1, /* calls to malloc/free */
	EXS_LOG_REG = 1 >> 2, /* calls to ibv_reg_mr/ibv_dereg_mr */
	EXS_LOG_SYNC = 1 >> 3, /* mutex, semaphore, and condition var ops */
	EXS_LOG_NEGO = 1 >> 4, /* negotiation sequence */
	EXS_LOG_SENDRECV = 1 >> 5, /* exs_send and exs_recv operations */
	EXS_LOG_QUEUE = 1 >> 6, /* event queue operations */
	EXS_LOG_STREAM = 1 >> 7, /* SOCK_STREAM operations */
	EXS_LOG_ACCEPT = 1 >> 8, /* exs_accept()-related operations */
};

#define TRACE_ERR(...) \
	exs_log(EXS_LOG_UNSPECIFIED, EXS_LOG_ERROR, \
			__FILE__, __LINE__, \
			__func__, __VA_ARGS__)

#define TRACE_INFO(module, ...) \
	exs_log(module, EXS_LOG_INFO, \
			__FILE__, __LINE__, \
				__func__, __VA_ARGS__)

#define OUR_PERROR(message, errcode) \
	do { \
		char errbuf[_MAX_STRERROR_MALLOC]; \
		if (strerror_r(errcode, errbuf, \
					_MAX_STRERROR_MALLOC)) { \
			snprintf(errbuf, _MAX_STRERROR_MALLOC, \
					"unknown"); \
		} \
		TRACE_ERR("%s: %s", message, errbuf); \
	} while (0)

#if defined(ALLOW_EXS_TRACE)

extern int _our_debug_flag;

#define TRACING_IS_ON	_our_debug_flag

#define TRACE_DBG(module, ...) \
	do	{ \
		if (_our_debug_flag) { \
			exs_log(module, EXS_LOG_DEBUG, \
					__FILE__, __LINE__, \
					__func__, __VA_ARGS__); \
		} \
	} while (0)

#define TRACE(...)						\
	do	{						\
		if(_our_debug_flag) {				\
			fprintf(stderr, __VA_ARGS__);		\
		}						\
	} while (0)

#define TRACE_VAR_DECLARE(type, name)	type name;

#define TRACE_VAR_ASSIGN(lhs, rhs)	lhs = rhs

#define TRACE_VAR_INCREMENT(var)	var ++

#define MY_NAME(string)						\
	static char	my_name[] = string;

#define PRINT_STATUS(...)					\
	do	{						\
		if(_our_debug_flag) {				\
			our_pstatus(__VA_ARGS__);		\
		}						\
	} while (0)

#define DUMP_BUF(...)						\
	do	{						\
		if(_our_debug_flag) {				\
			our_dump_buf(__VA_ARGS__);		\
		}						\
	} while (0)

#define TRACE_EVENT_TYPE(...)					\
	do	{						\
		if(_our_debug_flag) {				\
			our_trace_event_type(__VA_ARGS__);	\
		}						\
	} while (0)						\

/* Wrapper functions for pthread_mutex_lock and pthread_mutex_unlock that print
 * debugging information.  The first argument may be any arbitrary expression
 * that results in a struct or union type.  The second argument MUST be a struct
 * or union member name; it MUST NOT be an arbitrary expression. */
#define OUR_MEMBER_MUTEX_LOCK(ptr, mutex)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "LOCK " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); \
	pthread_mutex_lock(&(ptr)->mutex); \
	TRACE_DBG(EXS_LOG_SYNC, "PAST " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); } while (0);

#define OUR_MEMBER_MUTEX_UNLOCK(ptr, mutex)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); \
	pthread_mutex_unlock(&(ptr)->mutex); } while (0);

#define OUR_MEMBER_RWLOCK_RDLOCK(ptr, rwlock)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "RDLOCK " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); \
	pthread_rwlock_rdlock(&(ptr)->rwlock); \
	TRACE_DBG(EXS_LOG_SYNC, "PAST " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); } while (0);

#define OUR_MEMBER_RWLOCK_WRLOCK(ptr, rwlock)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "WRLOCK " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); \
	pthread_rwlock_wrlock(&(ptr)->rwlock); \
	TRACE_DBG(EXS_LOG_SYNC, "PAST " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); } while (0);

#define OUR_MEMBER_RWLOCK_UNLOCK(ptr, rwlock)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); \
	pthread_rwlock_unlock(&(ptr)->rwlock); } while (0);

#define OUR_MEMBER_SEM_WAIT(ptr, sem)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "WAIT " #ptr " %p " #sem " %p", \
		(ptr), &(ptr)->sem); \
	sem_wait(&(ptr)->sem); \
	TRACE_DBG(EXS_LOG_SYNC, "PAST " #ptr " %p " #sem " %p", \
		(ptr), &(ptr)->sem); } while (0);

#define OUR_MEMBER_SEM_POST(ptr, sem)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "POST " #ptr " %p " #sem " %p", \
		(ptr), &(ptr)->sem); \
	sem_post(&(ptr)->sem); } while (0);

/* TRYLOCK takes a third parameter which is a pointer to an integer in which to
 * place the result from pthread_mutex_trylock. */
#define OUR_MEMBER_MUTEX_TRYLOCK(ptr, mutex, ret)	do { \
	TRACE_DBG(EXS_LOG_SYNC, "TRYLOCK " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); \
	*(ret) = pthread_mutex_trylock(&(ptr)->mutex); \
	TRACE_DBG(EXS_LOG_SYNC, "PAST " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); } while (0);

#else

#define TRACING_IS_ON	0

#define TRACE_DBG(module, ...) do { } while (0)

#define TRACE(args...)						\
	do	{						\
	} while (0)

#define TRACE_VAR_DECLARE(type, name)

#define TRACE_VAR_ASSIGN(lhs, rhs)	do { } while (0)

#define TRACE_VAR_INCREMENT(var)	do { } while (0)

#define MY_NAME(string)

#define PRINT_STATUS(args...)					\
	do	{						\
	} while (0)

#define DUMP_BUF(args...)					\
	do	{						\
	} while (0)

#define TRACE_EVENT_TYPE(args...)				\
	do	{						\
	} while (0)

#define OUR_MEMBER_MUTEX_LOCK(ptr, mutex)	do { \
		pthread_mutex_lock(&(ptr)->mutex); \
	} while (0)

#define OUR_MEMBER_MUTEX_UNLOCK(ptr, mutex)	do { \
		pthread_mutex_unlock(&(ptr)->mutex); \
	} while (0)

#define OUR_MEMBER_RWLOCK_RDLOCK(ptr, rwlock)	do { \
		pthread_rwlock_rdlock(&(ptr)->rwlock); \
	} while (0)

#define OUR_MEMBER_RWLOCK_WRLOCK(ptr, rwlock)	do { \
		pthread_rwlock_wrlock(&(ptr)->rwlock); \
	} while (0)

#define OUR_MEMBER_RWLOCK_UNLOCK(ptr, rwlock)	do { \
		pthread_rwlock_unlock(&(ptr)->rwlock); \
	} while (0)

#define OUR_MEMBER_SEM_WAIT(ptr, sem)	do { \
		sem_wait(&(ptr)->sem); \
	} while (0)

#define OUR_MEMBER_SEM_POST(ptr, sem)	do { \
		sem_post(&(ptr)->sem); \
	} while (0)

/* TRYLOCK takes a third parameter which is a pointer to an integer in which to
 * place the result from pthread_mutex_trylock. */
#define OUR_MEMBER_MUTEX_TRYLOCK(ptr, mutex, ret)	do { \
	*(ret) = pthread_mutex_trylock(&(ptr)->mutex); } while (0)

#endif


#if defined(ALLOW_EXS_CALLOC)

#define EXS_CALLOC(nmemb, size)					\
	_our_calloc_calloc(nmemb, size, __FUNCTION__, __LINE__)

#define EXS_FREE(address)					\
	_our_calloc_free(address, __LINE__)

#define EXS_CALLOC_REPORT(fptr)						\
	_our_calloc_report(fptr)

extern void *
_our_calloc_calloc(size_t nmemb, size_t size, const char * fname,
		unsigned int lineno);

extern void
_our_calloc_free( void *address, unsigned int lineno );

extern void
_our_calloc_report( FILE *fptr );

#else

#define EXS_CALLOC(nmemb, size)					\
	calloc(nmemb, size)

#define EXS_FREE(address)					\
	free(address)

#define EXS_CALLOC_REPORT(fptr)					\
	do { } while(0)

#endif


#endif /* OUR_CALLOC_FREE_H_ */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
